import numpy as np

from .global_settings import DEBUG, ID_ADD, ID_MULT, UINT_DTYPE
from .helper_classes import PriorityQueue2D
from .helpers_fcts_numba import compile_valid_options, count_usage, factor_num_ops, num_ops_1D_horner, true_num_ops


class FactorisationNode(object):
    """
    A node representing a factorisation of a polynomial:
    p = f_1 * p_1 + p_2
    its sub problems are polynomials as well -> 'divide and conquer'
    """

    # prevent dynamic attribute assignment (-> safe memory)
    __slots__ = ['factor', 'node1_fact', 'node2', 'factorized_rows', 'non_factorized_rows', 'value_idx']

    def __init__(self, factor, node1_fact, node2, factorized_rows, non_factorized_rows):
        self.factor = factor
        self.node1_fact = node1_fact
        self.node2 = node2  # can be None if all monomials shared the factor
        self.factorized_rows = factorized_rows
        self.non_factorized_rows = non_factorized_rows
        self.value_idx = None

    def __str__(self):
        if type(self.factor) == tuple:
            dim, exp = self.factor
            if exp == 1:
                s = 'x_{}'.format(dim + 1)
            else:
                s = 'x_{}^{}'.format(dim + 1, exp)
        else:
            s = self.factor.__str__()
        s += ' (' + self.node1_fact.__str__() + ')'
        if self.node2 is not None:
            s += ' + ' + self.node2.__str__()
        return s

    def __gt__(self, other):
        # arbitrary, required for sorting in heap
        return True

    def compile_factors(self, coefficients, tree_coefficients, factor_container, id_counter):
        # create and store the unique ScalarFactor instance
        property_list = [self.factor]  # own factor is a scalar factor, only one property
        self.factor = factor_container.get_factor(property_list)

        coeffs1 = coefficients[self.factorized_rows]
        self.node1_fact.compile_factors(coeffs1, tree_coefficients, factor_container, id_counter)

        if self.node2 is not None:
            coeffs2 = coefficients[self.non_factorized_rows]
            self.node2.compile_factors(coeffs2, tree_coefficients, factor_container, id_counter)

        # this node does not need its own index in the value array, but can reuse the index of its first node
        self.value_idx = self.node1_fact.value_idxs[0]

    def get_recipe(self):
        # p = f_1 * p_1 + p_2
        # values[target] = values[target] *op* values[source]

        # IMPORTANT: sub trees have to be evaluated BEFORE their values can be used!
        # -> add their recipes first
        tree_recipe, op_recipe = self.node1_fact.get_recipe()

        # the value at values[node.idx] is the evaluated value of this node
        tree_recipe += [
            # instruction encoding: target, source
            # multiply the value of the node1 with the value of the factor
            (self.node1_fact.value_idxs[0], self.factor.value_idx),
        ]
        # separate: op (binary: 0/1)
        op_recipe += [ID_MULT]

        if self.node2 is not None:
            tree_recipe_sub, op_recipe_sub = self.node2.get_recipe()
            tree_recipe += tree_recipe_sub
            op_recipe += op_recipe_sub

            tree_recipe += [
                # add the value of node2 to this value
                (self.node1_fact.value_idxs[0], self.node2.value_idxs[0]),
            ]
            op_recipe += [ID_ADD]

        return tree_recipe, op_recipe


class OptimalFactorisationNode(FactorisationNode):
    # prevent dynamic attribute assignment (-> safe memory)
    __slots__ = ['cost_estimate', 'factorisation_measure', 'fully_factorized']

    def __init__(self, factor, node1_fact, node2, factorized_rows, non_factorized_rows):
        super(OptimalFactorisationNode, self).__init__(factor, node1_fact, node2, factorized_rows, non_factorized_rows)
        self.cost_estimate = 0
        self.factorisation_measure = 0
        self.fully_factorized = False
        self.update_properties()

    def update_properties(self):
        self.cost_estimate = factor_num_ops(*self.factor) + self.node1_fact.cost_estimate
        self.factorisation_measure = self.node1_fact.factorisation_measure

        if self.node2 is not None:
            self.cost_estimate += self.node2.cost_estimate
            self.factorisation_measure += self.node2.factorisation_measure
            self.fully_factorized = self.node1_fact.fully_factorized and self.node2.fully_factorized
        else:
            self.fully_factorized = self.node1_fact.fully_factorized

    def refine(self):
        assert not self.fully_factorized
        self.node1_fact.refine()
        if self.node2 is not None:
            self.node2.refine()
        self.update_properties()


class BasePolynomialNode(object):
    """
    The base class representing a multivariate polynomial as a child node in a factorisation tree for finding
        a good factorisation of a bigger multivariate polynomial
    """

    # prevent dynamic attribute assignment (-> safe memory)
    __slots__ = ['exponents', 'unique_exponents', 'num_monomials', 'dim', 'children', 'value_idxs', 'factors',
                 'has_children', 'children_class', 'factorisation_class']

    def __init__(self, exponents, *args, **kwargs):
        self.exponents = exponents
        self.dim = self.exponents.shape[1]
        self.num_monomials = self.exponents.shape[0]

        self.unique_exponents = []
        for d in range(self.dim):
            dim_exponents = self.exponents[:, d]
            # sort in ascending order (used for finding the most common scalar factor)
            dim_exponents_unique = np.unique(dim_exponents)
            dim_exponents_unique.sort()
            # exponent 0 is never a valid factorisation option
            if dim_exponents_unique[0] == 0:
                self.unique_exponents.append(dim_exponents_unique[1:])
            else:
                self.unique_exponents.append(dim_exponents_unique)

        self.value_idxs = None
        self.factors = None
        self.has_children = False
        self.children = None
        self.children_class = None
        self.factorisation_class = None
        self.post_init()

    def post_init(self):
        self.children_class = BasePolynomialNode
        self.factorisation_class = FactorisationNode

        # fully factorize automatically
        option = self.choose_option()
        if option is None:
            self.has_children = False
        else:
            self.has_children = True
            self.factorize(*option)

    def choose_option(self):
        """
        :return: the option with the maximum usage in all monomials
        """

        if self.num_monomials <= 1:
            # there are no options when only one monomial is left in a node!
            return None

        max_usage = 0
        max_usage_option = None

        for dim, dim_unique_exponents in enumerate(self.unique_exponents):
            # of all existing exponents, the smallest ones for sure have the highest usage
            for exp in dim_unique_exponents:
                usage = count_usage(dim, exp, self.exponents)
                if usage > max_usage:
                    max_usage = usage
                    max_usage_option = (dim, exp)
                # no other exponent in this dimension can have a higher usage
                break

        return max_usage_option

    def __str__(self):
        if self.has_children:
            return self.get_child().__str__()
        else:
            monomials = []
            for i, exp_vect in enumerate(self.exponents):
                monomial = ['c']
                for dim, exp in enumerate(exp_vect):
                    if exp > 0:
                        if exp == 1:
                            monomial.append('x_{}'.format(dim + 1))
                        else:
                            monomial.append('x_{}^{}'.format(dim + 1, exp))
                monomials.append(' '.join(monomial))
            return ' + '.join(monomials)

    def factorize(self, dim, exp):
        # create a representation of the polynomial with all the monomials 'benefiting' from factorisation
        factorized_rows = self.exponents[:, dim] >= exp
        exponents1_fact = self.exponents[factorized_rows, :]
        # the factor has to be deducted from the exponents (="factored out")
        exponents1_fact[:, dim] -= exp

        node1_fact = self.children_class(exponents1_fact)

        # create a representation of the polynomial with all the remaining monomials
        non_factorized_rows = np.invert(factorized_rows)
        if not np.any(non_factorized_rows):
            # no monomials remain, do not create a polynomial representation
            node2 = None
        else:
            exponents2 = self.exponents[non_factorized_rows, :]
            node2 = self.children_class(exponents2)
            if DEBUG:
                assert exp > 0
                assert not np.any(exponents1_fact < 0)
                assert not np.any(exponents2 < 0)
                assert exponents1_fact.shape[1] == exponents2.shape[1] and len(exponents2.shape) == 2
                assert exponents1_fact.shape[0] + exponents2.shape[0] == self.num_monomials

        factor = (dim, exp)
        child = self.factorisation_class(factor, node1_fact, node2, factorized_rows, non_factorized_rows)
        self.store_child(child)

    def store_child(self, child):
        self.children = child

    def get_child(self):
        return self.children

    def compile_factors(self, coefficients, tree_coefficients, factor_container, id_counter):
        """

        :param coefficients: numpy array of coefficients of the root polynomial
        :param tree_coefficients: list of coefficients in the order of appearance in the value array
        :param factor_container: class storing and managing all existing factors
        :param id_counter: counter object for assigning indices in the value array
        :return:
        """
        if self.has_children:
            child = self.get_child()
            child.compile_factors(coefficients, tree_coefficients, factor_container, id_counter)
            # this node does not need its own index in the value array, but can reuse the value index of its sub problem
            self.value_idxs = [child.value_idx]
        else:
            # this node cannot be factorized (no options)
            # for evaluation the sum of all evaluated monomials multiplied with their coefficients has to be computed
            # p = c1 * mon1 + c2 * mon2 ...
            self.factors = factor_container.compile_factors(self.exponents)
            # the coefficients are stored in the value array and all need their own index
            self.value_idxs = []
            for i, factor in enumerate(self.factors):
                self.value_idxs.append(id_counter.__next__())
                tree_coefficients.append(coefficients[i])
            # TODO
            assert (len(coefficients) == len(self.factors))

    def get_recipe(self):
        """
        :return: the list of instructions ("recipe") required for evaluating the represented factorisation
        """
        if self.has_children:
            # the own recipe is the recipe of its factorisation
            return self.get_child().get_recipe()
        else:
            # this node cannot be factorized
            # for evaluation the sum of all evaluated monomials multiplied with their coefficients has to be computed
            # p = c1 * mon1 + c2 * mon2 ...
            # the final evaluated value of this node has to be stored at the first of its value indices
            initial_coeff_value_idx = self.value_idxs[0]
            factor = self.factors[0]
            tree_recipe = []
            op_recipe = []

            if factor is not None:
                # multiply the first coefficients with the value of the first factor
                tree_recipe += [(initial_coeff_value_idx, self.factors[0].value_idx)]
                op_recipe += [ID_MULT]

            for i in range(1, len(self.value_idxs)):
                coeff_value_idx = self.value_idxs[i]
                factor = self.factors[i]
                if factor is not None:
                    # multiply each coefficients with the value of the factor
                    tree_recipe += [
                        (coeff_value_idx, factor.value_idx),
                    ]
                    op_recipe += [ID_MULT]

                # add this value to the previously computed value
                tree_recipe += [
                    (initial_coeff_value_idx, coeff_value_idx),
                ]
                op_recipe += [ID_ADD]
            return tree_recipe, op_recipe


class OptimalPolynomialNode(BasePolynomialNode):
    """
    A class representing a multivariate polynomial with all possible factorisation options.
    The different factorisation options are being stored sorted in a heap (self.children)
    An instance of this class is a node (=subproblem) in search tree for finding
    the optimal factorisation of a bigger multivariate polynomial
    """

    # prevent dynamic attribute assignment (-> safe memory)
    __slots__ = ['options', 'cost_estimate', 'factorisation_measure', 'fully_factorized']

    def __init__(self, exponents, *args, **kwargs):

        self.cost_estimate = 0
        self.factorisation_measure = 0
        self.fully_factorized = True
        self.options = None

        # self.post_init() is being called at the end
        super(OptimalPolynomialNode, self).__init__(exponents, *args, **kwargs)

    def post_init(self):
        self.children_class = OptimalPolynomialNode
        self.factorisation_class = OptimalFactorisationNode
        self.get_all_options()
        self.estimate_cost()

    def __eq__(self, other):
        """
        TODO improvement:
        different orders of factorisations can lead to identical sub problems
        solving them independently leads to unnecessary computations
        -> before creating a new node check if it already exists (equality comparison required)
        equality test can be aborted when certain global properties of nodes do not match
        (cost_estimate, dimensions of exponent matrix...)
        only actually compare exponent matrices for equality after all previous tests passed
        ATTENTION: after changing a multiply used node all parents have to be updated!
        """
        raise NotImplementedError

    def store_child(self, child):
        self.children.put(child)

    def get_child(self):
        # return the currently most promising factorisation
        return self.children.get()

    def estimate_cost(self):
        """
        :return: the estimated amount of operations required needed when evaluating this polynomial in fully
            factorized form. true amount when fully factorized otherwise a lower bound.

        Every existing factor (represented by exponents) has to be evaluated, which minimally
        requires the amount of operations all respective 1D-Horner factorisations would take.
        NOTE: not counting the amount of ADD operations and operations needed because of the coefficients,
            since they stay constant in sum irrespective of the factorisation (= number of monomials)

        TODO: improvement idea: more ops certainly needed when exponents appear multiple times in exponent vectors
         (monomials are unique) and can hence by definition not be fully factorized... but do not count multiple times
         -> derive rule.
        """
        if self.fully_factorized:
            # use the true number of operations
            self.cost_estimate = true_num_ops(self.exponents)
        else:
            heuristic = 0
            for dim_unique_exp in self.unique_exponents:
                heuristic += num_ops_1D_horner(dim_unique_exp)
            self.cost_estimate = heuristic

    def get_all_options(self):
        """
        :return: a list of all meaningful options (dim, exp) for factorisation

        NOTE: scalar factors (<-> exponents) which appear only in one monomial are not meaningful factors
            Picking them would only lead to separating this one monomial = equal to not factorising at all
            no decrease in #ops -> the heuristic would not change -> also no change in search
            monomials without any common exponents should not be factorized (can be computed without factorisation)
        """

        options = []
        if self.num_monomials <= 1:
            # there are no options when only one monomial is left in a node!
            return options

        for dim, dim_unique_exponents in enumerate(self.unique_exponents):
            usage_vector = np.zeros(dim_unique_exponents.shape, dtype=UINT_DTYPE)
            valid_option_vector = np.zeros(dim_unique_exponents.shape, dtype=bool)
            # TODO test
            valid_option_vector = compile_valid_options(dim, valid_option_vector, usage_vector, dim_unique_exponents,
                                                        self.exponents)

            for exp in dim_unique_exponents[valid_option_vector]:
                options.append((dim, exp))

        self.factorisation_measure = len(options)
        self.fully_factorized = len(options) == 0
        self.options = options
        pass

    def refine(self):
        """
        factorize further (grow tree)

        TODO improvement:
        initially compute good factorisation (with heuristic). use as upper bound for the #ops
        do not keep factorisations which have a higher or equal! estimated #ops (<- save memory)
        when heap becomes empty, the heuristic solution is optimal
        apply to all layers, pass upper bound down the tree (reduce accordingly)
        NOTE: when searching all optimal solutions having the same #ops as the upper bound is allowed!
        ATTENTION: cost estimate is not counting additions (cf. estimate_cost() ) do not compare with total #ops!

        TODO improvement:
        while loop keep factorising until no other option has a lower heuristic or fully factorised
        -> check what the next lowest heuristic is. corner case: no other option available!
        """
        # TODO remove
        # print('\nrefining', self.fully_factorized, self.cost_estimate, self)
        if self.fully_factorized:
            # this check is required, because nodes at the same depth in the factorisation tree might have different
            # factorisation statuses (.refine() is still being called)
            return

        if self.has_children:
            # the heap has been initialized
            # pick the most promising factorisation subtree, factorise it further
            promising_node = self.children.pop()  # pop, not get!
            # print('refining', promising_node.fully_factorized,  promising_node.cost_estimate, promising_node)
            promising_node.refine()
            # print('done', promising_node.fully_factorized,  promising_node.cost_estimate, promising_node)
            self.store_child(promising_node)
        else:
            # no factorisation has been done before
            # initialize heap
            # the cost estimate of less factorised bad factorisations and more factorised good factorisation is similar
            # to reach a fully factorised solution faster, more factorised options have to be favoured
            # -> have a "2D heap" sorted first after the cost estimate and then after the factorisation progress
            self.children = PriorityQueue2D()
            # use all options and store the different factorisation results in heap
            # there might be many options and trying all is often unnecessary,
            # but trying just one is also not efficient, because the unfactorized node most certainly
            # has a lower cost estimate and hence would get picked next anyway
            for dim, exp in self.options:
                self.factorize(dim, exp)

            self.has_children = True

        # print('all children:')
        # for c in self.children.get_all():
        #     print('\t', c.fully_factorized, c.cost_estimate,c)

        # the new properties of this node are equal to the properties of the most promising new subtree
        promising_node = self.get_child()
        # print('most promising:', promising_node.fully_factorized, promising_node.cost_estimate, promising_node)

        self.factorisation_measure = promising_node.factorisation_measure
        self.fully_factorized = promising_node.fully_factorized
        # cost_estimate_prev = self.cost_estimate
        self.cost_estimate = promising_node.cost_estimate
        # the initial heuristic is a lower bound for the true cost
        # the cost estimate should always increase or stay constant (= getting more realistic) with every refinement
        # assert cost_estimate_prev <= self.cost_estimate


class OptimalFactorisationRoot(OptimalPolynomialNode):
    """
    Class for finding optimal factorisations of given multivariate polynomials
    Basic idea: adapted A* search
        allow all meaningful possible factorisation while ranking them according to
        their lowest possible amount of operations needed for evaluation (heuristic)
        iteratively factorize the most promising factorisation further until a full factorisation has been found

    guaranteed to yield a factorisation with the minimal amount of operations required during evaluation (optimal)
    memory and time consumption is much higher than with using a heuristic to pick a single factorisation

    TODO improvement: analyse and optimize runtime of optimal factorisation search
    tradeoff between accuracy of the heuristic (-> ignore irrelevant factorisations)
        and time to compute heuristic (-> better to just try factorisations)


    TODO improvement idea: instead of just optimizing for the lowest amount of operations,
    also consider the different costs of the different operations on computer architectures
    (optimize total evaluation time).
    ATTENTION: a completely new heuristic for estimating the lower bound would be required!
    """

    def __init__(self, coefficients, tree_coefficients, id_counter, factor_container, *args, **kwargs):
        """
        :param coefficients: a numpy array column vector of doubles representing the coefficients of the monomials
        :param exponents: a numpy array matrix of unsigned integers representing the exponents of the monomials
            the ordering does not matter, but every exponent row has to be unique!
        :param rectify_input: whether to convert the input parameters into compatible numpy arrays
        :param validate_input: whether to check if the input parameters fulfill the requirements
        """
        super(OptimalFactorisationRoot, self).__init__(*args, **kwargs)

        while not self.fully_factorized:
            # TODO report current search status, verbosity option, not every loop
            # (#total factorisations done, #max depth, #open possibilities, upper bound, lower bound)
            self.refine()

        # the most promising factorisation (lowest cost estimate) is fully factorized now
        # collect all appearing factors in the best solution and assign ids (=idx in the value array) to the nodes
        self.compile_factors(coefficients, tree_coefficients, factor_container, id_counter)

    def find_all_optimal(self):
        """
        TODO refine and collect all fully factorized optimal solutions until the next best factorisation
            has a higher cost estimate
        Attention: do not remove entire subtree, but just one realisation each time!
        TODO analyse obtained optimal factorisations
        use this data to derive better heuristics for factorizing more optimally a priori
        :return:
        """
        raise NotImplementedError


class HeuristicFactorisationRoot(BasePolynomialNode):
    """
    Class for finding a good factorisation of a given multivariate polynomial by iteratively factoring out the most
    commonly used factor (heuristic).
    """

    def __init__(self, coefficients, tree_coefficients, id_counter, factor_container, *args, **kwargs):
        super(HeuristicFactorisationRoot, self).__init__(*args, **kwargs)
        # polynomial is fully factorized now
        # collect all appearing factors in the factorisation tree and assign ids (=idx in the value array) to the nodes
        self.compile_factors(coefficients, tree_coefficients, factor_container, id_counter)



import itertools

import numpy as np


# https://stackoverflow.com/questions/2068372/fastest-way-to-list-all-primes-below-n/3035188#3035188
# a generator yielding all prime numbers in ascending order
def erat2():
    D = {}
    yield 2
    for q in itertools.islice(itertools.count(3), 0, None, 2):
        p = D.pop(q, None)
        if p is None:
            D[q * q] = q
            yield q
        else:
            x = p + q
            while x in D or not (x & 1):
                x += p
            D[x] = p


def get_prime_array(length):
    return np.array(list(itertools.islice(erat2(), length)), dtype=np.uint32)


def get_goedel_id_of(prime_idx, exponent, prime_array):
    '''
    NOTE: factor IDs of monomials (product of scala IDs) potentially grow very large,
    especially with high dimensionality -> overflow?!
    python automatically uses the reqired long data type for ints,  no special attention required?!
    TODO use different method different from goedel id approach

    :param prime_idx:
    :param exponent:
    :param prime_array: the unique ID of any scalar monomial x_i^n
    :return:
    '''
    return int(prime_array[prime_idx] ** exponent)


def rectify(coefficients, exponents):
    """
    convert the input into numpy arrays valid as input to MultivarPolynomial
    raise an error if the given input is incompatible
    :param coefficients: possibly a python list of coefficients to be converted
    :param exponents: possibly a nested python list of exponents to be converted
    :return: the input converted into appropriate numpy data types
    """
    rectified_coefficients = np.atleast_1d(np.array(coefficients, dtype=np.float64)).reshape(-1, 1)

    rectified_exponents = np.atleast_2d(np.array(exponents, dtype=np.int))

    # exponents must not be negative!
    # ATTENTION: when converting to unsigned integer, negative integers become large!
    assert not np.any(rectified_exponents < 0)
    rectified_exponents = rectified_exponents.astype(np.uint32)

    # ignore the entries with 0.0 coefficients
    if np.any(rectified_coefficients == 0.0):
        non_zero_coeff_rows = np.where(rectified_coefficients != 0.0)[0]
        rectified_coefficients = rectified_coefficients[non_zero_coeff_rows, :]
        rectified_exponents = rectified_exponents[non_zero_coeff_rows, :]

    return rectified_coefficients, rectified_exponents


def validate(coefficients, exponents):
    """
    raises an error when the given input is not valid
    :param coefficients: a numpy array column vector of doubles representing the coefficients of the monomials
    :param exponents: a numpy array matrix of unsigned integers representing the exponents of the monomials
    :return:
    """
    # coefficients must be given as a column vector (2D)
    assert coefficients.shape[1] == 1 and len(coefficients.shape) == 2
    # exponents must be 2D (matrix) = a list of exponent vectors
    assert len(exponents.shape) == 2
    # there must be at least one entry
    assert coefficients.shape[0] > 0
    # exponents must not be negative
    assert not np.any(exponents < 0)
    # there must not be duplicate exponent vectors
    assert exponents.shape == np.unique(exponents, axis=0).shape
    # there must not be any coefficients with 0.0
    assert not np.any(coefficients == 0.0)
    # must have the same amount of entries
    assert coefficients.shape[0] == exponents.shape[0]



from numpy import float64, int64, uint32

ID_MULT = False
ID_ADD = True  # ATTENTION: is being used in helpers_fcts_numba.py/eval_recipe()

# numba is expecting certain data types (static typing):
INT_DTYPE = int64  # i8 =  8byte integer
UINT_DTYPE = uint32  # u4 =  4byte unsigned integer
FLOAT_DTYPE = float64  # f8 =  8byte float

DEFAULT_PICKLE_FILE_NAME = 'multivar_polynomial.pickle'

DEBUG = False
# DEBUG = True



import heapq  # implementation of the heap queue algorithm, also known as the priority queue algorithm (binary tree)

import numpy as np

from .helper_fcts import get_goedel_id_of


# modified sample code from https://www.redblobgames.com/pathfinding/a-star/
class PriorityQueue:
    def __init__(self):
        self.elements = []

    def __len__(self):
        return len(self.elements)

    def is_empty(self):
        return len(self.elements) == 0

    def put(self, item, cost):
        heapq.heappush(self.elements, (cost, item))

    def pop(self):
        # remove and returns (only) the item with the lowest cost
        return heapq.heappop(self.elements)

    def get(self):
        return self.elements[0]


class PriorityQueue2D:
    def __init__(self):
        self.lvl1_heap = PriorityQueue()
        self.cost2id = {}
        self.id2heap = {}

    def __len__(self):
        length = 0
        for lvl2_heap in self.id2heap.values():
            length += len(lvl2_heap)
        return length

    def is_empty(self):
        return self.lvl1_heap.is_empty()

    def put(self, item):
        cost1 = item.cost_estimate
        cost2 = item.factorisation_measure
        try:
            # look up the heap with this lvl1 cost
            heap_id = self.cost2id[cost1]
            lvl2_heap = self.id2heap[heap_id]
        except KeyError:
            # there exists no second level heap with this cost, create an empty one
            lvl2_heap = PriorityQueue()
            heap_id = id(lvl2_heap)
            self.cost2id[cost1] = heap_id
            self.id2heap[heap_id] = lvl2_heap
            # the lvl1 heap stores the ids of the corresponding heaps
            self.lvl1_heap.put(heap_id, cost1)

        # the lvl2 heaps store the actual items
        lvl2_heap.put(item, cost2)

    def pop(self):
        cost1, heap_id = self.lvl1_heap.get()
        lvl2_heap = self.id2heap[heap_id]
        cost2, item = lvl2_heap.pop()
        if lvl2_heap.is_empty():
            # remove entries referring to an empty heap
            self.lvl1_heap.pop()
            self.id2heap.pop(heap_id)
            self.cost2id.pop(cost1)
            del lvl2_heap

        return item

    def get(self):
        # try:
        cost1, heap_id = self.lvl1_heap.get()
        lvl2_heap = self.id2heap[heap_id]
        cost2, item = lvl2_heap.get()
        return item  # return only the item without the costs
        # except IndexError:
        #     return None

    def get_all(self):
        all_items = []
        for lvl2_heap in self.id2heap.values():
            for (cost, item) in lvl2_heap.elements:
                all_items.append(item)
        return all_items


class AbstractFactor(object):
    __slots__ = ['monomial_id', 'value_idx']

    def __str__(self):
        raise NotImplementedError()

    def __repr__(self):
        raise NotImplementedError()

    def eval(self, factor_values):
        # the value of all used factors has already been computed
        return factor_values[self.value_idx]

    def compute(self, x, factor_values):
        raise NotImplementedError()


class ScalarFactor(AbstractFactor):
    # a monomial with just one variable: x_i^n
    __slots__ = ['dimension', 'exponent']

    def __init__(self, factor_dimension, factor_exponent, factor_id):
        self.dimension = factor_dimension
        self.exponent = factor_exponent
        self.monomial_id = factor_id
        self.value_idx = None  # initialize the idx with None to catch faulty evaluation tries

    def __str__(self):
        # variable numbering starts with 1: x_1, x_2, ...
        if self.exponent == 1:
            return 'x_{}'.format(self.dimension + 1)
        else:
            return 'x_{}^{}'.format(self.dimension + 1, self.exponent)

    def __repr__(self):
        return self.__str__()

    @staticmethod
    def num_ops(self):
        # count the number of instructions done during compute (during eval() only looks up the computed value)
        return 1

    def compute(self, x, factor_values):
        factor_values[self.value_idx] = x[self.dimension] ** self.exponent

    def get_recipe(self):
        """
        for evaluation the input value has to be either copied or exponentiated
        :return: copy_recipe, scalar_recipe, monomial_recipe
        """
        if self.exponent == 1:
            # just copy x[dim] value
            # values[target] = x[source]
            return [(self.value_idx, self.dimension)], [], []
        else:
            # instruction encoding: target, source, exponent
            # values[target] = x[source] ** exponent
            return [], [(self.value_idx, self.dimension, self.exponent)], []


class MonomialFactor(AbstractFactor):
    """
    factorisation_idxs: the indices of the values of all factors.
        cannot be set at construction time, because the list of all factors is sorted
        at the end of building the horner factorisation tree (when all factors are known)
    """
    # a monomial consisting of a product of scalar factors: x_i^j * x_k^l * ...
    __slots__ = ['scalar_factors', 'factorisation_idxs']

    def __init__(self, scalar_factors):
        self.scalar_factors = scalar_factors

        # the id of a monomial is the Goedel number of their exponent vector
        # https://en.wikipedia.org/wiki/G%C3%B6del_numbering
        # this is a unique id for every monomial
        # the id of a monomial the product of the ids of its scalar factors
        self.monomial_id = 1
        for f in self.scalar_factors:
            self.monomial_id *= f.monomial_id

        if len(self.scalar_factors) == 1:
            raise ValueError('There is only one factor. Use ScalarFactor instead')

        if len(self.scalar_factors) == 0:
            raise ValueError()

        self.value_idx = None  # initialize the idx with None to catch faulty evaluation tries

    def __str__(self):
        return ' '.join([f.__str__() for f in self.scalar_factors])

    def __repr__(self):
        return self.__str__()

    def num_ops(self):
        # count the number of instructions done during compute (during eval() only looks up the computed value)
        return len(self.factorisation_idxs) - 1

    def compute(self, x, factor_values):
        # IMPORTANT: compute() of all the sub factors has had to be called before!
        value = factor_values[self.factorisation_idxs[0]]
        for idx in self.factorisation_idxs[1:]:
            value *= factor_values[idx]

        factor_values[self.value_idx] = value

    def get_recipe(self):
        """
        for evaluation all scalar factors have to be multiplied (= monomial_recipe)
        :return: copy_recipe, scalar_recipe, monomial_recipe
        """
        # target = source1 * source2
        # instruction encoding: target, source1, source2
        target, source1, source2 = self.value_idx, self.factorisation_idxs[0], self.factorisation_idxs[1]
        monomial_recipe = [(target, source1, source2)]

        source1 = target  # always take the previously computed value
        for source2 in self.factorisation_idxs[2:]:
            # and multiply it with the remaining factor values
            monomial_recipe += [(target, source1, source2)]

        return [], [], monomial_recipe


class FactorContainer:
    def __init__(self, prime_array):
        self.factors = []
        self.id2idx = {}
        self.prime_array = prime_array

    def get_factor(self, property_list):

        scalar_factors = []
        monomial_id = 1
        for d, e in property_list:
            scalar_id = get_goedel_id_of(d, e, self.prime_array)
            try:
                scalar_idx = self.id2idx[scalar_id]
                scalar_factor = self.factors[scalar_idx]
            except KeyError:
                scalar_factor = ScalarFactor(d, e, scalar_id)
                self.id2idx[scalar_id] = len(self.factors)
                self.factors.append(scalar_factor)

            scalar_factors.append(scalar_factor)
            monomial_id *= scalar_id

        assert len(scalar_factors) > 0

        if len(scalar_factors) == 1:
            # the requested factor is scalar factor
            return scalar_factors[0]

        # the requested factor is a monomial consisting of multiple factors
        try:
            monomial_idx = self.id2idx[monomial_id]
            monomial_factor = self.factors[monomial_idx]
        except KeyError:
            monomial_factor = MonomialFactor(scalar_factors)
            self.id2idx[monomial_id] = len(self.factors)
            self.factors.append(monomial_factor)
        return monomial_factor

    def compile_factors(self, exponent_matrix):
        # all monomials are independent
        if np.all(np.sum(exponent_matrix, axis=0) != np.max(exponent_matrix, axis=0)):
            print(exponent_matrix)

        all_factors = []
        for mon_nr in range(exponent_matrix.shape[0]):
            exponent_vector = exponent_matrix[mon_nr, :]
            scalar_factors = []
            property_list = []
            for dim, exp in enumerate(exponent_vector):
                if exp > 0:
                    prop = (dim, exp)
                    property_list.append(prop)
                    scalar_factor = self.get_factor([prop])
                    scalar_factors.append(scalar_factor)

            if len(scalar_factors) == 0:
                factor = None
            elif len(scalar_factors) == 1:
                factor = scalar_factors[0]
            else:
                factor = self.get_factor(property_list)

            all_factors.append(factor)

        return all_factors



# -*- coding:utf-8 -*-

import os
import re
import sys

from setuptools import setup


def get_version(package):
    """
    Return package version as listed in `__version__` in `__init__.py`.
    """
    init_py = open(os.path.join(package, '__init__.py')).read()
    return re.search("__version__ = ['\"]([^'\"]+)['\"]", init_py).group(1)


version = get_version('multivar_horner')

with open('README.rst') as f:
    readme = f.read()

with open('CHANGELOG.rst') as changelog_file:
    changelog = changelog_file.read()

# https://stackoverflow.com/questions/23174738/setup-py-packages-and-unicode-literals
native_string_pckg_name = 'multivar_horner'
if sys.version_info.major == 2:
    native_string_pckg_name = b'multivar_horner'

setup(
    name='multivar_horner',
    version=version,
    packages=['multivar_horner'],
    # package_data={
    #     native_string_pckg_name: [],
    # },
    description='python package implementing a multivariate horner scheme for efficiently'
                ' evaluating multivariate polynomials',
    author='J. Michelfeit',
    author_email='python@michelfe.it',
    license='MIT licence',
    url='https://github.com/MrMinimal64/multivar_horner',  # use the URL to the github repo
    keywords='math mathematics polynomial polynomials polynomial-evaluation multivariate multivariate-polynomials'
             ' horner horner-scheme factorisation factorization',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Intended Audience :: Information Technology',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.0',
        'Programming Language :: Python :: 3.1',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Mathematics',
        'Topic :: Scientific/Engineering :: Physics',
    ],
    long_description=readme + '\n\n' + changelog,
    install_requires=[
        'numpy',
        'numba',
    ],
)



# TODO
# for Ahead-Of-Time Compilation:
# from numba.pycc import CC
# cc = CC('compiled_helpers', )
# # Uncomment the following line to print out the compilation steps
# cc.verbose = True


# precompiled time critical helper functions
import numpy as np
from numba import b1, f8, jit, u4


# TODO     TypingError: numba doesn't support kwarg for prod
# @jit(f8(f8[:], f8[:], u4[:, :]), nopython=True, cache=True)
def naive_eval(x, coefficients, exponents):
    return np.sum(coefficients.T * np.prod(np.power(x, exponents), axis=1), axis=1)[0]


# @cc.export('eval_compiled', 'f8(f8[:], f8[:], u4[:, :], u4[:, :], u4[:, :])')
@jit(f8(f8[:], f8[:], u4[:, :], u4[:, :], u4[:, :], u4[:, :], b1[:]), nopython=True, cache=True)
def eval_recipe(x, value_array, copy_recipe, scalar_recipe, monomial_recipe, tree_recipe, tree_ops):
    # IMPORTANT: the order of following the recipes is not arbitrary!
    # print(value_array)

    # copy recipe instruction encoding: target, source
    # [target, source] = copy_recipe[i, :]
    for i in range(copy_recipe.shape[0]):
        # value_array[target] = x[source1]
        value_array[copy_recipe[i, 0]] = x[copy_recipe[i, 1]]

    # print('computing scalar factors: ...')
    # scalar recipe instruction encoding: target, source, exponent
    # [target, source1, exponent] = scalar_recipe[i, :]
    for i in range(scalar_recipe.shape[0]):
        # print('value[{}] = {} ^ {}'.format(target, x[source1], exponent))
        # value_array[target] = x[source1] ** exponent
        value_array[scalar_recipe[i, 0]] = x[scalar_recipe[i, 1]] ** scalar_recipe[i, 2]

    # print('computing monomial factors: ...')
    # monomial recipe instruction encoding: target, source1, source2
    # [target, source1, source2] = monomial_recipe[i, :]
    for i in range(monomial_recipe.shape[0]):
        # print('value[{}] = {} * {} (idx: {}, {})'.format(target, value_array[source1], value_array[source2], source1,
        #                                                 source2))
        # value_array[target] = value_array[source1] * value_array[source2]
        value_array[monomial_recipe[i, 0]] = value_array[monomial_recipe[i, 1]] * value_array[monomial_recipe[i, 2]]

    # print('evaluating factorisation tree: ...')
    # tree recipe instruction encoding: target, source
    # [target, source] = tree_recipe[i, :]
    # separate operation array: *op_id*
    # value_array[target] = value_array[target] *op* value_array[source]
    for i in range(tree_recipe.shape[0]):
        target = tree_recipe[i, 0]
        if tree_ops[i]:  # ADDITION: 1
            # print('value[{}] = {} + {}'.format(target, value_array[target], value_array[source1]))
            # value_array[target] = value_array[target] + value_array[source1]
            value_array[target] = value_array[target] + value_array[tree_recipe[i, 1]]
        else:
            # print('value[{}] = {} * {}'.format(target, value_array[target], value_array[source1]))
            # value_array[target] = value_array[target] * value_array[source1]
            value_array[target] = value_array[target] * value_array[tree_recipe[i, 1]]

    # the value at the first position is the value of the polynomial
    return value_array[0]


@jit(u4(u4[:]), nopython=True, cache=True)
def num_ops_1D_horner(unique_exponents):
    """
    :param unique_exponents: np array of unique exponents sorted in increasing order without 0
    :return: the number of operations of the one dimensional horner factorisation
        without counting additions (just MUL & POW) and without considering the coefficients


    NOTE: in 1D the horner factorisation is both unique and optimal (minimal amount of operations)
    """
    nr_unique_exponents = unique_exponents.shape[0]
    # the exponent 0 is not present!
    assert not np.any(unique_exponents == 0)

    if nr_unique_exponents == 0:
        return 0

    # one MUL operation is required !between! all factors in the factorisation chain
    # the amount of factors (= "length of factorisation chain") is equal to the amount of unique existing exponents
    num_ops = nr_unique_exponents - 1

    # start with exponent 0 (not in unique exponents)
    # the difference between one and the next exponent determines if a POW operation is needed to evaluate a factor
    # unique exponents are sorted: prev_exp < exp
    prev_exp = 0
    for i in range(nr_unique_exponents):
        exp = unique_exponents[i]
        if exp - prev_exp >= 2:
            num_ops += 1  # 1 POW operation
        prev_exp = exp

    return num_ops


@jit(u4(u4[:, :]), nopython=True, cache=True)
def true_num_ops(exponent_matrix):
    """
    without counting additions (just MUL & POW) and but WITH considering the coefficients (1 MUL per monomial)
    """
    num_ops = 0
    for monomial_nr in range(exponent_matrix.shape[0]):
        for dim in range(exponent_matrix.shape[1]):
            exp = exponent_matrix[monomial_nr, dim]
            if exp > 0:
                # per scalar factor 1 MUL operation is required
                num_ops += 1
                if exp >= 2:
                    # for scalar factors with exponent >= 2 additionally 1 POW operation is required
                    num_ops += 1

    return num_ops


@jit(u4[:](u4, u4[:], u4[:], u4[:, :]), nopython=True, cache=True)
def compile_usage(dim, usage_vector, unique_exponents, exponent_matrix):
    """
    :return: a vector with the usage count of every unique exponent
    """

    for i in range(exponent_matrix.shape[0]):
        exp = exponent_matrix[i, dim]
        for j in range(len(unique_exponents)):
            if exp < unique_exponents[j]:
                break
            usage_vector[j] += 1

    return usage_vector


@jit(b1[:](u4, b1[:], u4[:], u4[:], u4[:, :]), nopython=True, cache=True)
def compile_valid_options(dim, valid_option_vector, usage_vector, unique_exponents, exponent_matrix):
    if len(valid_option_vector) == 0:
        # there are no unique exponents
        return valid_option_vector

    usage_vector = compile_usage(dim, usage_vector, unique_exponents, exponent_matrix)

    for exp_idx in range(usage_vector.size):
        # stop at the highest exponent having a usage >=2
        if usage_vector[exp_idx] >= 2:
            valid_option_vector[exp_idx] = True
        else:
            # all higher exponents have a lower usage!
            break

    return valid_option_vector


@jit(u4(u4, u4, u4[:, :]), nopython=True, cache=True)
def count_usage(dim, exp, exponent_matrix):
    """
    :return: the amount of times a scalar factor appears in the monomials
    """

    usage_cnt = 0
    for i in range(exponent_matrix.shape[0]):
        if exponent_matrix[i, dim] >= exp:
            usage_cnt += 1

    return usage_cnt


@jit(u4(u4, u4), nopython=True, cache=True)
def factor_num_ops(dim, exp):
    """
    :param factor: a tuple (dim, exp) representing the scalar factor: x_dim^exp
    :return: the amount of operations required to evaluate the given scalar factor
    """
    if exp >= 2:
        # 1 MUL + 1 POW
        return 2
    else:
        # 1 MUL
        return 1



# TODO s:

# matlab binding?
# optimize naive version without factorisation: compute all needed exponents only once. order exponents cleverly.
# store at which index the later entries (exponents in column) will just be 0 -> reduce the matrix size piecewise

# test gradient
# multivariate newton raphson method
# mention new features in readme
# function to predict the factorisation time and efficiency (time saved) based on dim, max_degree and num_entries
# based on system
# suggest function to use

# test factorisation for univariate polynomials! should always find optimum (unique 1D horner fact.)


import itertools
import pickle

import numpy as np

from .factorisation_classes import HeuristicFactorisationRoot, OptimalFactorisationRoot
from .global_settings import DEBUG, DEFAULT_PICKLE_FILE_NAME, FLOAT_DTYPE, UINT_DTYPE
from .helper_classes import FactorContainer, ScalarFactor
from .helper_fcts import get_prime_array, rectify, validate
from .helpers_fcts_numba import eval_recipe, naive_eval


# is not a helper function to make it an importable part of the package
def load_pickle(path=DEFAULT_PICKLE_FILE_NAME):
    print('importing polynomial from file "{}"'.format(path))
    with open(path, 'rb') as f:
        return pickle.load(f)


class MultivarPolynomial(object):
    """
    naive representation of a multivariate polynomial without any horner factorisation

    dim: the dimensionality of the polynomial.
    order: (for a multivariate polynomial) maximum sum of exponents in any of its monomials
    max_degree: the largest exponent in any of its monomials
        NOTE: the polynomial actually needs not to depend on all dimensions
    unused_variables: the dimensions the polynomial does not depend on
    """

    # prevent dynamic attribute assignment (-> safe memory)
    __slots__ = ['coefficients', 'exponents', 'existing_exponents', 'num_monomials', 'dim', 'order', 'max_degree',
                 'unused_variables', 'representation', 'heuristic']

    def __init__(self, coefficients, exponents, rectify_input=False, validate_input=False):
        """
        :param coefficients: a numpy array column vector of doubles representing the coefficients of the monomials
        :param exponents: a numpy array matrix of unsigned integers representing the exponents of the monomials
            the ordering does not matter, but every exponent row has to be unique!
        :param rectify_input: whether to convert the input parameters into compatible numpy arrays
        :param validate_input: whether to check if the input parameters fulfill the requirements
        """

        if rectify_input:
            coefficients, exponents = rectify(coefficients, exponents)

        if validate_input:
            validate(coefficients, exponents)

        self.coefficients = coefficients
        self.exponents = exponents

        self.num_monomials = self.exponents.shape[0]
        self.dim = self.exponents.shape[1]
        self.order = np.sum(self.exponents, axis=0).max()
        self.max_degree = self.exponents.max()

        self.unused_variables = np.where(~np.any(self.exponents, axis=1))[0]
        self.representation = ''
        self.existing_exponents = None
        self.heuristic = None

    def __str__(self):
        return self.get_representation()

    def __repr__(self):
        return self.__str__()

    def get_representation(self):

        s = '[#ops={}] p(x) = '.format(self.get_num_ops())
        monomials = []
        for i, exp_vect in enumerate(self.exponents):
            monomial = [str(self.coefficients[i, 0])]
            for dim, exp in enumerate(exp_vect):
                # if exp > 0:
                # show all operations, even 1 * x_i^0
                monomial.append('x_{}^{}'.format(dim + 1, exp))
            monomials.append(' '.join(monomial))

        s += ' + '.join(monomials)
        return s

    def export_pickle(self, path=DEFAULT_PICKLE_FILE_NAME):
        print('storing polynomial in file "{}" ...'.format(path))
        with open(path, 'wb') as f:
            pickle.dump(self, f)
        print('...done.\n')

    def get_num_ops(self):
        # count the number of instructions done when evaluating polynomial:
        y, x = self.exponents.shape
        # exponentiation: x*y
        # multiplication coefficient and scalar factors (monomials): x*y
        # sum: y-1
        return 2 * x * y + y - 1

    def eval(self, x, validate_input=True):
        """
        :param x:
        :param validate_input:
        :return:
        """

        if validate_input:
            x = np.array(x)
            assert len(x.shape) == 1
            assert x.shape[0] == self.dim

        return naive_eval(x, self.coefficients, self.exponents)

    def get_partial_derivative(self, i):
        """
        TODO test
        f(x) = a x^b
        f'(x) = ab x^(b-1)
        :param i:
        :return: the partial derivative of this polynomial wrt. the i-th coordinate
        # ATTENTION: coordinate counting starts with 1! (first coordinates is #1)
        """

        assert (0 < i <= self.dim)
        coord_index = i - 1
        # IMPORTANT: do not modify the stored coefficient and exponent arrays of self!
        # set all the coefficients not depending on the i-th coordinate to 0
        # this simply means not adding them to the list of coefficients of the new polynomial class instance
        active_rows = np.where(self.exponents[:, coord_index] >= 1)[0]

        new_coefficients = self.coefficients[active_rows]
        new_exponents = self.exponents[active_rows, :]

        if DEBUG:
            assert new_coefficients.shape[0] == new_exponents.shape[0]
            assert new_coefficients.shape[1] == 1 and len(new_coefficients.shape) == 2
            assert new_exponents.shape[1] == self.dim and len(new_exponents.shape) == 2

        # multiply the coefficients with the exponent of the i-th coordinate
        new_coefficients = np.multiply(new_coefficients.flatten(), new_exponents[:, coord_index])
        new_coefficients = new_coefficients.reshape(-1, 1)

        # reduce the the exponent of the i-th coordinate by 1
        new_exponents[:, coord_index] -= 1

        # must call the proper constructor method also for inherited classes
        # TODO also use option 'univariate_factors'
        return self.__class__(new_coefficients, new_exponents)

    def get_gradient(self):
        """
        :return: the list of all partial derivatives
        """
        return [self.get_partial_derivative(i) for i in range(1, self.dim + 1)]


class HornerMultivarPolynomial(MultivarPolynomial):
    """
    a representation of a multivariate polynomial using horner factorisation to save evaluation time

    dim: the dimensionality of the polynomial.
    order: (for a multivariate polynomial) maximum sum of exponents in any of its monomials
    max_degree: the largest exponent in any of its monomials
        NOTE: the polynomial actually needs not to depend on all dimensions
    unused_variables: the dimensions the polynomial does not depend on
    """
    # __slots__ declared in parents are available in child classes. However, child subclasses will get a __dict__
    # and __weakref__ unless they also define __slots__ (which should only contain names of any additional slots).
    __slots__ = ['prime_array', 'factorisation_tree', 'factor_container', 'num_ops', 'initial_value_array',
                 'copy_recipe', 'scalar_recipe', 'monomial_recipe', 'tree_recipe', 'tree_ops']

    def __init__(self, coefficients, exponents, rectify_input=False, validate_input=False, keep_tree=False,
                 compute_representation=False, find_optimal=False):
        """
        :param coefficients: a numpy array column vector of doubles representing the coefficients of the monomials
        :param exponents: a numpy array matrix of unsigned integers representing the exponents of the monomials
            the ordering does not matter, but every exponent row has to be unique!
        :param rectify_input: whether to convert the input parameters into compatible numpy arrays
        :param validate_input: whether to check if the input parameters fulfill the requirements
        """

        def compute_num_ops():
            # count the number of instructions done during evaluation
            # (computing all factors + evaluating the factorisation subtrees)
            # for every entry in each recipe one operation is being done
            # do not count "copy" instructions (not really a mathematical operation)
            self.num_ops = self.scalar_recipe.shape[0] + self.monomial_recipe.shape[0] + self.tree_ops.shape[0]

        def get_string_representation():
            representation = '[#ops={}] p(x)'.format(self.num_ops)
            if compute_representation:
                representation += ' = ' + self.factorisation_tree.__str__()
            return representation

        super(HornerMultivarPolynomial, self).__init__(coefficients, exponents, rectify_input, validate_input)

        # the needed prime numbers for computing all goedel numbers of all used factors
        self.prime_array = get_prime_array(self.dim)

        # store all unique factors of the horner factorisation
        # NOTE: do NOT create all scalar factors with exponent 1 (they might be unused!)
        self.factor_container = FactorContainer(self.prime_array)

        # factorize the polynomial once and store the factorisation as a tree
        id_counter = itertools.count(0)
        tree_coefficients = []
        if find_optimal:
            self.factorisation_tree = OptimalFactorisationRoot(self.coefficients, tree_coefficients, id_counter,
                                                               self.factor_container, exponents=self.exponents)
        else:
            self.factorisation_tree = HeuristicFactorisationRoot(self.coefficients, tree_coefficients, id_counter,
                                                                 self.factor_container, exponents=self.exponents)

        # factor list is now filled with the unique factors
        # during evaluation of a polynomial the values of all the unique factors are needed at least once
        # -> compute the values of all factors (monomials) once and store them
        # store a pointer to the computed value for every unique factor
        # save instructions for the evaluation -> find factorisation of all factors = again a factorisation tree
        # sort and factorize the monomials to quickly evaluate them once during a query
        self.link_monomials()

        # compile and store a "recipe" for evaluating the polynomial with just numpy arrays
        self.initial_value_array, self.copy_recipe, self.scalar_recipe, self.monomial_recipe, self.tree_recipe, \
            self.tree_ops = self.compile_recipes(tree_coefficients)

        compute_num_ops()
        self.representation = get_string_representation()

        if not keep_tree:
            # the trees and factors are not needed any more
            # a value lookup can be done with just the recipe
            # free up the memory
            del self.factorisation_tree
            del self.factor_container
            del self.prime_array

    def __str__(self):
        return self.representation

    def get_num_ops(self):
        return self.num_ops

    def link_monomials(self):
        """
        find the optimal factorisation of the unique factors themselves
        since the monomial ids are products of the ids of their scalar factors
        check for the highest possible divisor among all factor ids
        this leads to a minimal factorisation for quick evaluation of the monomial values
        :return:
        """
        # sort after their id
        unique_factor_id_list = list(sorted(self.factor_container.id2idx.keys()))
        unique_factors = list(sorted(self.factor_container.factors, key=lambda f: f.monomial_id))
        # property of the list: the scalar factors of each monomial are stored in front of it

        # IMPORTANT: properly set the indices of the values for each factor
        # the values of every factor are stored after the coefficients
        value_idx_offset = self.num_monomials
        for idx, factor in enumerate(unique_factors):
            factor.value_idx = idx + value_idx_offset

        # start at the last factor (highest id)
        pointer1 = len(unique_factors) - 1

        # the smallest factor has no factorisation (stop at pointer=1)
        while pointer1 > 0:
            candidate = unique_factors[pointer1]

            if type(candidate) is ScalarFactor:
                # scalar factors have no useful factorisation
                # (can be directly computed with the same amount of operations)
                pointer1 -= 1
                continue

            # print(1, self.unique_factors[pointer1])
            remaining_factor_id = candidate.monomial_id
            # store the indices of the factors that divide the id
            factorisation_idxs = []

            pointer2 = pointer1 - 1
            # find the factors with the highest id which are a factor of the current monomial
            # TODO back track when no factorisation is possible with that factor
            # try with the remaining
            while 1:
                if pointer2 < 0:
                    # no factorisation of this monomial has been found, because the remainder after
                    # picking a factorizing monomial cannot be factorised itself
                    # just pick the scalar factors of the monomial
                    # ATTENTION: offset needed!
                    factorisation_idxs = [unique_factor_id_list.index(scalar_factor.monomial_id) + value_idx_offset
                                          for scalar_factor
                                          in candidate.scalar_factors]
                    break

                monomial_id2 = unique_factors[pointer2].monomial_id

                if remaining_factor_id < monomial_id2:
                    # this monomial has a higher id than the remaining factor. it cannot be a factorisation
                    pointer2 -= 1
                    continue

                if monomial_id2 == remaining_factor_id:
                    # the last factor has been found
                    # ATTENTION: offset needed!
                    factorisation_idxs.append(pointer2 + value_idx_offset)
                    break

                quotient, remainder = divmod(remaining_factor_id, monomial_id2)
                if remainder == 0:
                    # this factor is a factor of the monomial
                    # ATTENTION: offset needed!
                    factorisation_idxs.append(pointer2 + value_idx_offset)
                    # reduce the id
                    remaining_factor_id = quotient

                pointer2 -= 1

            # print('found factorisation: {} ='.format(self.unique_factors[pointer1]),
            #       ' * '.join([str(self.unique_factors[idx]) for idx in reversed(factorisation_idxs)]))
            # store the indices of the factors that factorize the monomial
            #   to quickly access their values when computing the value of the monomial
            candidate.factorisation_idxs = factorisation_idxs

            pointer1 -= 1

    def compile_recipes(self, tree_coefficients):
        # compile a recipe encoding all needed instructions in order to evaluate the polynomial
        # = clever data structure for representing the factorisation tree
        # -> avoid recursion and function call overhead while evaluating

        # for evaluating the polynomial in tree form intermediary computation results have to be stored in a value array
        # the value array has one entry for every subtree and one for every factor
        value_array_length = self.num_monomials + len(self.factor_container.factors)

        initial_value_array = np.empty(value_array_length, dtype=FLOAT_DTYPE)

        # initial_value_array = np.array(tree_coefficients, dtype=FLOAT_DTYPE)..reshape(value_array_length)
        # the initial value array has the coefficients of all subtrees stored at the index of their id
        for i, v in enumerate(tree_coefficients):
            initial_value_array[i] = v

        # self.horner_tree.fill_value_array(value_array)

        # compile the recipes for computing the factors
        # scalar factors (depending on x) are being evaluated differently
        #   from the monomial factors (depending on scalar factors)
        copy_recipe = []  # skip computing factors with exp 1, just copy x vals
        scalar_recipe = []
        monomial_recipe = []
        for f in self.factor_container.factors:
            copy_instr, scalar_instr, monomial_instr = f.get_recipe()
            copy_recipe += copy_instr
            scalar_recipe += scalar_instr
            monomial_recipe += monomial_instr

        # compile the recipe for evaluating the horner factorisation tree
        tree_recipe, tree_ops = self.factorisation_tree.get_recipe()
        # convert and store the recipes
        # for the recipes numba is expecting the data types:
        #   array(uint, 2d, C),
        #   separate boolean array for operations, uint not needed (just 0 or 1)
        return (initial_value_array,
                np.array(copy_recipe, dtype=UINT_DTYPE).reshape((-1, 2)),
                np.array(scalar_recipe, dtype=UINT_DTYPE).reshape((-1, 3)),
                np.array(monomial_recipe, dtype=UINT_DTYPE).reshape((-1, 3)),
                np.array(tree_recipe, dtype=UINT_DTYPE).reshape((-1, 2)),
                np.array(tree_ops, dtype=np.bool))

    def eval(self, x, validate_input=False):
        """
        make use of numba precompiled evaluation function:
        IDEA: encode factorisation in numpy array. "recipe" which values to add or multiply when
        TODO parallel computing, split up the recipe in independent parts (evaluation of sub trees)
        :param x:
        :param validate_input: whether to check if the input parameters fulfill the requirements
        :return:
        """

        if validate_input:
            x = np.array(x)
            assert len(x.shape) == 1
            assert x.shape[0] == self.dim

        # IMPORTANT: copy the initial value array
        # the array is being used as temporal storage and values would get the overwritten
        return eval_recipe(x, self.initial_value_array.copy(), self.copy_recipe, self.scalar_recipe,
                           self.monomial_recipe, self.tree_recipe, self.tree_ops)



import random
import timeit
import unittest
from math import log10

import numpy as np
import pytest

from multivar_horner.global_settings import UINT_DTYPE
from multivar_horner.multivar_horner import HornerMultivarPolynomial, MultivarPolynomial

MAX_DIM = 2
MAX_DEGREE = 2
NR_SAMPLES = 200

poly_settings_list = []
input_list = []
poly_class_instances = []


# TODO automatically also test only_scalar_factor=True
# TODO compare difference in computed values (error)
# TODO test consecutive evaluations of the same polynomial!

def proto_test_case(data, fct):
    all_good = True
    for input, expected_output in data:
        # print(input, expected_output, fct(input))
        actual_output = fct(input)
        if actual_output != expected_output:
            print('input: {} expected: {} got: {}'.format(input, expected_output, actual_output))
            all_good = False

    assert all_good


#
# def id2exponent_vect(prime_list, monomial_id):
#     # find the exponent vector corresponding to a monomial id
#     # = prime decomposition
#     exponent_vect = np.zeros(prime_list.shape, dtype=UINT_DTYPE)
#     current_id = monomial_id
#     for i, prime in enumerate(prime_list):
#         while 1:
#             quotient, remainder = divmod(current_id, prime)
#             if remainder == 0:
#                 exponent_vect[i] += 1
#                 current_id = quotient
#             else:
#                 break
#
#         if current_id == 0:
#             break
#
#     if current_id != 0:
#         raise ValueError('no factorisation found')
#
#     return exponent_vect
#
#
# def _sparse_range_generator(max_value, density):
#     for i in range(max_value):
#         if random.random() < density:
#             yield i


def random_polynomial_settings(all_exponents, max_abs_coeff=1.0):
    # every exponent can take the values in the range [0; max_degree]
    max_nr_exponent_vects = all_exponents.shape[0]

    # decide how many entries the polynomial should have
    # desired for meaningful speed test results:
    # every possible polynomial should appear with equal probability
    # there must be at least 1 entry
    nr_exponent_vects = random.randint(1, max_nr_exponent_vects)

    row_idxs = list(range(max_nr_exponent_vects))
    assert max_nr_exponent_vects >= nr_exponent_vects
    for length in range(max_nr_exponent_vects, nr_exponent_vects, -1):
        # delete random entry from list
        row_idxs.pop(random.randint(0, length - 1))

    assert len(row_idxs) == nr_exponent_vects

    exponents = all_exponents[row_idxs, :]
    coefficients = (np.random.rand(nr_exponent_vects, 1) - 0.5) * (2 * max_abs_coeff)
    return coefficients, exponents


def all_possible_exponents(dim, max_degree):
    def cntr2exp_vect(cntr):
        exp_vect = np.empty((dim), dtype=UINT_DTYPE)
        for d in range(dim - 1, -1, -1):
            divisor = (max_degree + 1) ** d
            # cntr = quotient*divisor + remainder
            quotient, remainder = divmod(cntr, divisor)
            exp_vect[d] = quotient
            cntr = remainder
        return exp_vect

    max_nr_exponent_vects = (max_degree + 1) ** dim
    all_possible = np.empty((max_nr_exponent_vects, dim), dtype=UINT_DTYPE)
    for i in range(max_nr_exponent_vects):
        # print(i, cntr2exp_vect(i))
        all_possible[i] = cntr2exp_vect(i)

    return all_possible


def rnd_settings_list(length, dim, max_degree):
    all_exponent_vect = all_possible_exponents(dim, max_degree)
    settings_list = [random_polynomial_settings(all_exponent_vect) for i in range(length)]

    # # random settings should have approx. half the amount of maximal entries on average
    # num_monomial_entries = 0
    # for settings in settings_list:
    #     num_monomial_entries += settings[0].shape[0]
    #
    # avg_monomial_entries = num_monomial_entries / length
    # max_monomial_entries = int((max_degree + 1) ** dim)
    # print(avg_monomial_entries, max_monomial_entries)
    return settings_list


def rnd_input_list(length, dim, max_abs_val):
    return [(np.random.rand(dim) - 0.5) * (2 * max_abs_val) for i in range(length)]


def setup_time_fct(poly_class):
    # store instances globally to directly use them for eval time test
    global poly_settings_list, poly_class_instances

    poly_class_instances = []
    for settings in poly_settings_list:
        poly_class_instances.append(poly_class(*settings))


def eval_time_fct():
    global poly_class_instances, input_list

    for instance, input in zip(poly_class_instances, input_list):
        instance.eval(input)


def get_avg_num_ops():
    global poly_class_instances

    num_ops = 0
    for instance in poly_class_instances:
        num_ops += instance.get_num_ops()

    avg_num_ops = round(num_ops / len(poly_class_instances))
    return avg_num_ops


def time_preprocess(time):
    valid_digits = 4
    zero_digits = abs(min(0, int(log10(time))))
    digits_to_print = zero_digits + valid_digits
    return str(round(time, digits_to_print))


def difference(quantity1, quantity):
    speedup = round((quantity / quantity1 - 1), 1)
    if speedup < 0:
        speedup = round((quantity1 / quantity - 1), 1)
        if speedup > 10.0:
            speedup = round(speedup)

        return str(speedup) + ' x less'
    else:
        if speedup > 10.0:
            speedup = round(speedup)
        return str(speedup) + ' x more'


def compute_lucrativity(setup_horner, setup_naive, eval_horner, eval_naive):
    benefit_eval = eval_naive - eval_horner
    loss_setup = setup_horner - setup_naive
    # x * benefit = loss
    return round(loss_setup / benefit_eval)


def speed_test_run(dim, max_degree, nr_samples, template):
    global poly_settings_list, input_list, poly_class_instances

    poly_settings_list = rnd_settings_list(nr_samples, dim, max_degree)
    input_list = rnd_input_list(nr_samples, dim, max_abs_val=1.0)

    setup_time_naive = timeit.timeit("setup_time_fct(MultivarPolynomial)", globals=globals(), number=1)
    setup_time_naive = setup_time_naive / NR_SAMPLES  # = avg. per sample

    # poly_class_instances is not populated with the naive polynomial class instances
    # print(poly_class_instances[0])
    num_ops_naive = get_avg_num_ops()

    eval_time_naive = timeit.timeit("eval_time_fct()", globals=globals(), number=1)
    eval_time_naive = eval_time_naive / NR_SAMPLES  # = avg. per sample

    setup_time_horner = timeit.timeit("setup_time_fct(HornerMultivarPolynomial)", globals=globals(), number=1)
    setup_time_horner = setup_time_horner / NR_SAMPLES  # = avg. per sample

    # poly_class_instances is not populated with the horner polynomial class instances
    # print(poly_class_instances[0])
    num_ops_horner = get_avg_num_ops()

    eval_time_horner = timeit.timeit("eval_time_fct()", globals=globals(), number=1)
    eval_time_horner = eval_time_horner / NR_SAMPLES  # = avg. per sample

    setup_delta = difference(setup_time_naive, setup_time_horner)
    eval_delta = difference(eval_time_naive, eval_time_horner)
    ops_delta = difference(num_ops_naive, num_ops_horner)
    lucrative_after = compute_lucrativity(setup_time_horner, setup_time_naive, eval_time_horner, eval_time_naive)

    print(template.format(str(dim), str(max_degree), time_preprocess(setup_time_naive),
                          time_preprocess(setup_time_horner), str(setup_delta),
                          time_preprocess(eval_time_naive), time_preprocess(eval_time_horner),
                          str(eval_delta), str(num_ops_naive), str(num_ops_horner), ops_delta, str(lucrative_after)))


class MainTest(unittest.TestCase):

    def test_eval(self):
        def cmp_value_fct(inp):
            print()
            coeff, exp, x = inp
            x = np.array(x).T
            poly = MultivarPolynomial(coeff, exp, rectify_input=True, validate_input=True)
            res1 = poly.eval(x, validate_input=True)
            print(str(poly))

            horner_poly = HornerMultivarPolynomial(coeff, exp, rectify_input=True, validate_input=True,
                                                   compute_representation=True, find_optimal=False)
            res2 = horner_poly.eval(x, validate_input=True)
            print(str(horner_poly))
            # print('x=',x.tolist())

            horner_poly_opt = HornerMultivarPolynomial(coeff, exp, rectify_input=True, validate_input=True,
                                                       compute_representation=True, find_optimal=True)
            res3 = horner_poly_opt.eval(x, validate_input=True)
            print(str(horner_poly_opt))
            # print('x=',x.tolist())

            if res1 != res2 or res2 != res3:
                print('results differ:', res1, res2, res3)
                assert False

            assert horner_poly.num_ops >= horner_poly_opt.num_ops

            return poly.eval(x, validate_input=True)

        invalid_test_data = [
            # calling with x of another dimension
            (([1.0, 2.0, 3.0],
              [[3, 1, 0], [2, 0, 1], [1, 1, 1]],
              [-2.0, 3.0]),
             # p(x) = 5.0 + 2.0* (-2)^1 + 1.0* (-2)^2 + 2.0* (-2)^2 *3^1 = 5.0 + 2.0* (-2) + 1.0* 4 + 2.0* 4 *3
             29.0),

            (([1.0, 2.0, 3.0],
              [[3, 1, 0], [2, 0, 1], [1, 1, 1]],
              [-2.0, 3.0, 1.0, 4.0]),
             # p(x) = 5.0 + 2.0* (-2)^1 + 1.0* (-2)^2 + 2.0* (-2)^2 *3^1 = 5.0 + 2.0* (-2) + 1.0* 4 + 2.0* 4 *3
             29.0),

            # negative exponents are not allowed
            (([1.0, 2.0, 3.0],
              [[3, -1, 0], [2, 0, 1], [1, 1, 1]],
              [-2.0, 3.0, 1.0, 4.0]),
             # p(x) = 5.0 + 2.0* (-2)^1 + 1.0* (-2)^2 + 2.0* (-2)^2 *3^1 = 5.0 + 2.0* (-2) + 1.0* 4 + 2.0* 4 *3
             29.0),

            # duplicate exponent entries are not allowed
            # negative exponents are not allowed
            (([1.0, 2.0, 3.0],
              [[3, 1, 0], [3, 1, 0], [2, 0, 1], [1, 1, 1]],
              [-2.0, 3.0, 1.0, 4.0]),
             # p(x) = 5.0 + 2.0* (-2)^1 + 1.0* (-2)^2 + 2.0* (-2)^2 *3^1 = 5.0 + 2.0* (-2) + 1.0* 4 + 2.0* 4 *3
             29.0),

        ]

        for inp, expected_output in invalid_test_data:
            with pytest.raises(AssertionError):
                cmp_value_fct(inp)

        invalid_test_data = [
            #
            # p(x) =  5.0
            (([5.0],  # coefficients
              [0],  # exponents
              [0.0]),  # x
             5.0),  # p(x)

            # p(1.0) = 1.0
            (([5.0],
              [0],
              [1.0]),
             5.0),

            # p(-1.0) = -1.0
            (([5.0],
              [0],
              [-1.0]),
             5.0),

            # p(33.5) =33.5
            (([5.0],
              [0],
              [33.5]),
             5.0),

            # p(x) =  1.0* x_1^1
            # p(0.0) = 0.0
            (([1.0],  # coefficients
              [1],  # exponents
              [0.0]),  # x
             0.0),  # p(x)

            # p(1.0) = 1.0
            (([1.0],
              [1],
              [1.0]),
             1.0),

            # p(-1.0) = -1.0
            (([1.0],
              [1],
              [-1.0]),
             -1.0),

            # p(33.5) =33.5
            (([1.0],
              [1],
              [33.5]),
             33.5),

            # p(x) =  1.0* x_1^1 + 0.0* * x_2^1
            (([1.0, 0.0],
              [[1, 0], [0, 1]],
              [0.0, 0.0]),
             0.0),

            (([1.0, 0.0],
              [[1, 0], [0, 1]],
              [1.0, 0.0]),
             1.0),

            (([1.0, 0.0],
              [[1, 0], [0, 1]],
              [-1.0, 0.0]),
             -1.0),

            (([1.0, 0.0],
              [[1, 0], [0, 1]],
              [33.5, 0.0]),
             33.5),

            # p(x) =  5.0 +  1.0* x_1^1
            (([5.0, 1.0],
              [[0, 0], [1, 0]],
              [0.0, 0.0]),
             5.0),

            (([5.0, 1.0],
              [[0, 0], [1, 0]],
              [1.0, 0.0]),
             6.0),

            (([5.0, 1.0],
              [[0, 0], [1, 0]],
              [-1.0, 0.0]),
             4.0),

            (([5.0, 1.0],
              [[0, 0], [1, 0]],
              [33.5, 0.0]),
             38.5),

            # p(x) =  5.0 + 2.0* x_1^1 + 1.0* x_1^2
            (([5.0, 2.0, 1.0],
              [[0, 0], [1, 0], [2, 0]],
              [0.0, 0.0]),
             5.0),

            (([5.0, 2.0, 1.0],
              [[0, 0], [1, 0], [2, 0]],
              [1.0, 0.0]),
             8.0),  # p(x) =  5.0 + 2.0 + 1.0

            (([5.0, 2.0, 1.0],
              [[0, 0], [1, 0], [2, 0]],
              [-1.0, 0.0]),
             4.0),  # p(x) =  5.0 - 2.0 + 1.0

            (([5.0, 2.0, 1.0],
              [[0, 0], [1, 0], [2, 0]],
              [2.0, 0.0]),
             13.0),  # p(x) =  5.0 + 2.0* 2.0^1 + 1.0* 2.0^2

            # p(x) =  5.0 + 2.0* x_1^1 + 1.0* x_1^2 + 2.0* x_1^2 *x_2^1
            (([5.0, 2.0, 1.0, 2.0],
              [[0, 0], [1, 0], [2, 0], [2, 1]],
              [0.0, 0.0]),
             5.0),

            (([5.0, 2.0, 1.0, 2.0],
              [[0, 0], [1, 0], [2, 0], [2, 1]],
              [1.0, 0.0]),
             8.0),  # p(x) =  5.0 + 2.0* 1^1 + 1.0* 1^2 + 2.0* 1^2 *0^1

            (([5.0, 2.0, 1.0, 2.0],
              [[0, 0], [1, 0], [2, 0], [2, 1]],
              [1.0, 1.0]),
             10.0),  # p(x) =  5.0 + 2.0* 1^1 + 1.0* 1^2 + 2.0* 1^2 *1^1

            (([5.0, 2.0, 1.0, 2.0],
              [[0, 0], [1, 0], [2, 0], [2, 1]],
              [-1.0, 0.0]),
             4.0),  # p(x) =  5.0 + 2.0* (-1)^1 + 1.0* (-1)^2 + 2.0* (-1)^2 *0^1

            (([5.0, 2.0, 1.0, 2.0],
              [[0, 0], [1, 0], [2, 0], [2, 1]],
              [-1.0, 1.0]),
             6.0),  # p(x) =  5.0 + 2.0* (-1)^1 + 1.0* (-1)^2 + 2.0* (-1)^2 *1^1

            (([5.0, 2.0, 1.0, 2.0],
              [[0, 0], [1, 0], [2, 0], [2, 1]],
              [-1.0, 2.0]),
             8.0),  # p(x) =  5.0 + 2.0* (-1)^1 + 1.0* (-1)^2 + 2.0* (-1)^2 *2^1

            (([5.0, 2.0, 1.0, 2.0],
              [[0, 0], [1, 0], [2, 0], [2, 1]],
              [-1.0, 3.0]),
             10.0),  # p(x) =  5.0 + 2.0* (-1)^1 + 1.0* (-1)^2 + 2.0* (-1)^2 *3^1

            (([5.0, 2.0, 1.0, 2.0],
              [[0, 0], [1, 0], [2, 0], [2, 1]],
              [-2.0, 3.0]),
             # p(x) = 5.0 + 2.0* (-2)^1 + 1.0* (-2)^2 + 2.0* (-2)^2 *3^1 = 5.0 + 2.0* (-2) + 1.0* 4 + 2.0* 4 *3
             29.0),

            # [20] p(x) = 1.0 x_1^3 x_2^1 + 2.0 x_1^2 x_3^1 + 3.0 x_1^1 x_2^1 x_3^1
            # [17] p(x) = x_2^1 [ x_1^3 [ 1.0 ] + x_1^1 x_3^1 [ 3.0 ] ] + x_1^2 x_3^1 [ 2.0 ]
            (([1.0, 2.0, 3.0],
              [[3, 1, 0], [2, 0, 1], [1, 1, 1]],
              [-2.0, 3.0, 1.0]),
             -34.0),

            # [27] p(x) = 1.0 x_3^1 + 2.0 x_1^3 x_2^3 + 3.0 x_1^2 x_2^3 x_3^1 + 4.0 x_1^1 x_2^5 x_3^1
            (([1.0, 2.0, 3.0, 4.0],
              [[0, 0, 1], [3, 3, 0], [2, 3, 1], [1, 5, 1]],
              [-2.0, 3.0, 1.0]),
             -2051.0),
        ]

        proto_test_case(invalid_test_data, cmp_value_fct)

    def test_speed(self):
        # this test also fulfills the purpose of testing the robustness of the code
        # (many random polygons are being created and evaluated)

        # todo write results to file
        # TODO test for larger dimensions
        # TODO plot speed
        # TODO compare & plot the performance wrt. the "density" of the polynomials
        # naive should stay constant, horner should get slower

        print('\nSpeed test:')
        print('testing {} evenly distributed random polynomials'.format(NR_SAMPLES))
        print('average timings per polynomial:\n')

        print(' {0:11s}  |  {1:38s} |  {2:35s} |  {3:35s} | {4:20s}'.format('parameters', 'setup time (/s)',
                                                                            'eval time (/s)',
                                                                            '# operations', 'lucrative after '))
        template = '{0:3s} | {1:7s} | {2:10s} | {3:10s} | {4:13s} | {5:10s} | {6:10s} | {7:10s} | {8:10s} | ' \
                   '{9:10s} | {10:10s} | {11:10s}'

        print(template.format('dim', 'max_deg', 'naive', 'horner', 'delta', 'naive',
                              'horner', 'delta', 'naive', 'horner', 'delta', '   # evals'))
        print('=' * 160)

        for dim in range(1, MAX_DIM + 1):
            for max_degree in range(1, MAX_DEGREE + 1):
                # TODO make num samples dependent on parameters, print
                speed_test_run(dim, max_degree, NR_SAMPLES, template)
            print()


if __name__ == '__main__':
    suite = unittest.TestLoader().loadTestsFromTestCase(MainTest)
    unittest.TextTestRunner(verbosity=2).run(suite)
    unittest.main()
